@@ -10,21 +10,30 @@ from contextlib import (
     redirect_stdout,
 )
 from typing import (
+    IO,
+    TYPE_CHECKING,
+    Any,
+    List,
+    NamedTuple,
     Optional,
+    TextIO,
+    Union,
+    cast,
 )
 
-from .utils import (
+from .utils import (  # namedtuple_with_defaults,
     StdSim,
-    namedtuple_with_defaults,
 )
 
+if TYPE_CHECKING:  # pragma: no cover
+    import cmd2
 
-class CommandResult(namedtuple_with_defaults('CommandResult', ['stdout', 'stderr', 'stop', 'data'])):
+
+class CommandResult(NamedTuple):
     """Encapsulates the results from a cmd2 app command
 
     :stdout: str - output captured from stdout while this command is executing
     :stderr: str - output captured from stderr while this command is executing
-             None if no error captured.
     :stop: bool - return value of onecmd_plus_hooks after it runs the given
            command line.
     :data: possible data populated by the command.
@@ -55,6 +64,12 @@ class CommandResult(namedtuple_with_defaults('CommandResult', ['stdout', 'stderr
        Named tuples are immutable. The contents are there for access,
        not for modification.
     """
+
+    stdout: str = ''
+    stderr: str = ''
+    stop: bool = False
+    data: Any = None
+
     def __bool__(self) -> bool:
         """Returns True if the command succeeded, otherwise False"""
 
@@ -69,16 +84,17 @@ class CommandResult(namedtuple_with_defaults('CommandResult', ['stdout', 'stderr
 
 class PyBridge:
     """Provides a Python API wrapper for application commands."""
-    def __init__(self, cmd2_app):
+
+    def __init__(self, cmd2_app: 'cmd2.Cmd') -> None:
         self._cmd2_app = cmd2_app
         self.cmd_echo = False
 
         # Tells if any of the commands run via __call__ returned True for stop
         self.stop = False
 
-    def __dir__(self):
+    def __dir__(self) -> List[str]:
         """Return a custom set of attribute names"""
-        attributes = []
+        attributes: List[str] = []
         attributes.insert(0, 'cmd_echo')
         return attributes
 
@@ -95,7 +111,7 @@ class PyBridge:
             echo = self.cmd_echo
 
         # This will be used to capture _cmd2_app.stdout and sys.stdout
-        copy_cmd_stdout = StdSim(self._cmd2_app.stdout, echo=echo)
+        copy_cmd_stdout = StdSim(cast(Union[TextIO, StdSim], self._cmd2_app.stdout), echo=echo)
 
         # Pause the storing of stdout until onecmd_plus_hooks enables it
         copy_cmd_stdout.pause_storage = True
@@ -107,18 +123,20 @@ class PyBridge:
 
         stop = False
         try:
-            self._cmd2_app.stdout = copy_cmd_stdout
-            with redirect_stdout(copy_cmd_stdout):
-                with redirect_stderr(copy_stderr):
+            self._cmd2_app.stdout = cast(TextIO, copy_cmd_stdout)
+            with redirect_stdout(cast(IO[str], copy_cmd_stdout)):
+                with redirect_stderr(cast(IO[str], copy_stderr)):
                     stop = self._cmd2_app.onecmd_plus_hooks(command, py_bridge_call=True)
         finally:
             with self._cmd2_app.sigint_protection:
-                self._cmd2_app.stdout = copy_cmd_stdout.inner_stream
+                self._cmd2_app.stdout = cast(IO[str], copy_cmd_stdout.inner_stream)
                 self.stop = stop or self.stop
 
-        # Save the output. If stderr is empty, set it to None.
-        result = CommandResult(stdout=copy_cmd_stdout.getvalue(),
-                               stderr=copy_stderr.getvalue() if copy_stderr.getvalue() else None,
-                               stop=stop,
-                               data=self._cmd2_app.last_result)
+        # Save the result
+        result = CommandResult(
+            stdout=copy_cmd_stdout.getvalue(),
+            stderr=copy_stderr.getvalue(),
+            stop=stop,
+            data=self._cmd2_app.last_result,
+        )
         return result
