@@ -16,13 +16,14 @@ from enum import (
 )
 from typing import (
     Any,
+    Deque,
     Optional,
     Sequence,
     Tuple,
     Union,
 )
 
-from wcwidth import (
+from wcwidth import (  # type: ignore[import]
     wcwidth,
 )
 
@@ -32,24 +33,6 @@ from . import (
     utils,
 )
 
-# This is needed for compatibility with early versions of Python 3.5 prior to 3.5.4
-try:
-    from typing import Deque
-except ImportError:  # pragma: no cover
-    import typing
-
-    # The following copied from the implementation of Deque in Python 3.5.4
-    # noinspection PyProtectedMember, PyUnresolvedReferences
-    class Deque(deque, typing.MutableSequence[typing.T]):
-        __slots__ = ()
-        __extra__ = deque
-
-        def __new__(cls, *args, **kwds):
-            if typing._geqv(cls, Deque):
-                raise TypeError('Type Deque cannot be instantiated; use deque() instead')
-            return typing._generic_new(deque, cls, *args, **kwds)
-
-
 # Constants
 EMPTY = ''
 SPACE = ' '
@@ -57,6 +40,7 @@ SPACE = ' '
 
 class HorizontalAlignment(Enum):
     """Horizontal alignment of text in a cell"""
+
     LEFT = 1
     CENTER = 2
     RIGHT = 3
@@ -64,6 +48,7 @@ class HorizontalAlignment(Enum):
 
 class VerticalAlignment(Enum):
     """Vertical alignment of text in a cell"""
+
     TOP = 1
     MIDDLE = 2
     BOTTOM = 3
@@ -71,12 +56,18 @@ class VerticalAlignment(Enum):
 
 class Column:
     """Table column configuration"""
-    def __init__(self, header: str, *, width: Optional[int] = None,
-                 header_horiz_align: HorizontalAlignment = HorizontalAlignment.LEFT,
-                 header_vert_align: VerticalAlignment = VerticalAlignment.BOTTOM,
-                 data_horiz_align: HorizontalAlignment = HorizontalAlignment.LEFT,
-                 data_vert_align: VerticalAlignment = VerticalAlignment.TOP,
-                 max_data_lines: Union[int, float] = constants.INFINITY) -> None:
+
+    def __init__(
+        self,
+        header: str,
+        *,
+        width: Optional[int] = None,
+        header_horiz_align: HorizontalAlignment = HorizontalAlignment.LEFT,
+        header_vert_align: VerticalAlignment = VerticalAlignment.BOTTOM,
+        data_horiz_align: HorizontalAlignment = HorizontalAlignment.LEFT,
+        data_vert_align: VerticalAlignment = VerticalAlignment.TOP,
+        max_data_lines: Union[int, float] = constants.INFINITY,
+    ) -> None:
         """
         Column initializer
 
@@ -98,7 +89,7 @@ class Column:
         if width is not None and width < 1:
             raise ValueError("Column width cannot be less than 1")
         else:
-            self.width = width
+            self.width: int = width if width is not None else -1
 
         self.header_horiz_align = header_horiz_align
         self.header_vert_align = header_vert_align
@@ -125,6 +116,7 @@ class TableCreator:
     implement a more granular API rather than use TableCreator directly. There are ready-to-use examples of this
     defined after this class.
     """
+
     def __init__(self, cols: Sequence[Column], *, tab_width: int = 4) -> None:
         """
         TableCreator initializer
@@ -146,14 +138,11 @@ class TableCreator:
 
             # For headers with the width not yet set, use the width of the
             # widest line in the header or 1 if the header has no width
-            if col.width is None:
-                line_widths = [ansi.style_aware_wcswidth(line) for line in col.header.splitlines()]
-                line_widths.append(1)
-                col.width = max(line_widths)
+            if col.width <= 0:
+                col.width = max(1, ansi.widest_line(col.header))
 
     @staticmethod
-    def _wrap_long_word(word: str, max_width: int, max_lines: Union[int, float],
-                        is_last_word: bool) -> Tuple[str, int, int]:
+    def _wrap_long_word(word: str, max_width: int, max_lines: Union[int, float], is_last_word: bool) -> Tuple[str, int, int]:
         """
         Used by _wrap_text() to wrap a long word over multiple lines
 
@@ -228,7 +217,12 @@ class TableCreator:
         :param max_lines: maximum lines to wrap before ending the last line displayed with an ellipsis
         :return: wrapped text
         """
-        def add_word(word_to_add: str, is_last_word: bool):
+
+        # MyPy Issue #7057 documents regression requiring nonlocals to be defined earlier
+        cur_line_width = 0
+        total_lines = 0
+
+        def add_word(word_to_add: str, is_last_word: bool) -> None:
             """
             Called from loop to add a word to the wrapped text
 
@@ -258,10 +252,9 @@ class TableCreator:
                         room_to_add = False
 
                 if room_to_add:
-                    wrapped_word, lines_used, cur_line_width = TableCreator._wrap_long_word(word_to_add,
-                                                                                            max_width,
-                                                                                            max_lines - total_lines + 1,
-                                                                                            is_last_word)
+                    wrapped_word, lines_used, cur_line_width = TableCreator._wrap_long_word(
+                        word_to_add, max_width, max_lines - total_lines + 1, is_last_word
+                    )
                     # Write the word to the buffer
                     wrapped_buf.write(wrapped_word)
                     total_lines += lines_used - 1
@@ -318,6 +311,11 @@ class TableCreator:
             if data_line_index > 0:
                 wrapped_buf.write('\n')
 
+            # If the last line is empty, then add a newline and stop
+            if data_line_index == len(data_str_lines) - 1 and not data_line:
+                wrapped_buf.write('\n')
+                break
+
             # Locate the styles in this line
             styles = utils.get_styles_in_text(data_line)
 
@@ -370,8 +368,7 @@ class TableCreator:
 
         return wrapped_buf.getvalue()
 
-    def _generate_cell_lines(self, cell_data: Any, is_header: bool,
-                             col: Column, fill_char: str) -> Tuple[Deque[str], int]:
+    def _generate_cell_lines(self, cell_data: Any, is_header: bool, col: Column, fill_char: str) -> Tuple[Deque[str], int]:
         """
         Generate the lines of a table cell
 
@@ -401,11 +398,18 @@ class TableCreator:
         aligned_text = utils.align_text(wrapped_text, fill_char=fill_char, width=col.width, alignment=text_alignment)
 
         lines = deque(aligned_text.splitlines())
-        cell_width = max([ansi.style_aware_wcswidth(line) for line in lines])
+        cell_width = ansi.widest_line(aligned_text)
         return lines, cell_width
 
-    def generate_row(self, *, row_data: Optional[Sequence[Any]] = None, fill_char: str = SPACE,
-                     pre_line: str = EMPTY, inter_cell: str = (2 * SPACE), post_line: str = EMPTY) -> str:
+    def generate_row(
+        self,
+        *,
+        row_data: Optional[Sequence[Any]] = None,
+        fill_char: str = SPACE,
+        pre_line: str = EMPTY,
+        inter_cell: str = (2 * SPACE),
+        post_line: str = EMPTY,
+    ) -> str:
         """
         Generate a header or data table row
 
@@ -425,11 +429,13 @@ class TableCreator:
         :raises: ValueError if fill_char, pre_line, inter_cell, or post_line contains an unprintable
                  character like a newline
         """
+
         class Cell:
             """Inner class which represents a table cell"""
+
             def __init__(self) -> None:
                 # Data in this cell split into individual lines
-                self.lines = []
+                self.lines: Deque[str] = deque()
 
                 # Display width of this cell
                 self.width = 0
@@ -453,11 +459,10 @@ class TableCreator:
             raise TypeError("Fill character must be exactly one character long")
 
         # Look for unprintable characters
-        validation_dict = {'fill_char': fill_char, 'pre_line': pre_line,
-                           'inter_cell': inter_cell, 'post_line': post_line}
+        validation_dict = {'fill_char': fill_char, 'pre_line': pre_line, 'inter_cell': inter_cell, 'post_line': post_line}
         for key, val in validation_dict.items():
             if ansi.style_aware_wcswidth(val) == -1:
-                raise (ValueError("{} contains an unprintable character".format(key)))
+                raise ValueError(f"{key} contains an unprintable character")
 
         # Number of lines this row uses
         total_lines = 0
@@ -529,40 +534,49 @@ class SimpleTable(TableCreator):
     Implementation of TableCreator which generates a borderless table with an optional divider row after the header.
     This class can be used to create the whole table at once or one row at a time.
     """
-    # Spaces between cells
-    INTER_CELL = 2 * SPACE
 
-    def __init__(self, cols: Sequence[Column], *, tab_width: int = 4, divider_char: Optional[str] = '-') -> None:
+    def __init__(
+        self, cols: Sequence[Column], *, column_spacing: int = 2, tab_width: int = 4, divider_char: Optional[str] = '-'
+    ) -> None:
         """
         SimpleTable initializer
 
         :param cols: column definitions for this table
+        :param column_spacing: how many spaces to place between columns. Defaults to 2.
         :param tab_width: all tabs will be replaced with this many spaces. If a row's fill_char is a tab,
                           then it will be converted to one space.
         :param divider_char: optional character used to build the header divider row. Set this to None if you don't
                              want a divider row. Defaults to dash. (Cannot be a line breaking character)
-        :raises: TypeError if fill_char is more than one character (not including ANSI style sequences)
-        :raises: ValueError if text or fill_char contains an unprintable character
+        :raises: ValueError if column_spacing is less than 0
+        :raises: ValueError if tab_width is less than 1
+        :raises: TypeError if divider_char is longer than one character
+        :raises: ValueError if divider_char is an unprintable character
         """
+        if column_spacing < 0:
+            raise ValueError("Column spacing cannot be less than 0")
+        self.inter_cell = column_spacing * SPACE
+
         if divider_char is not None:
             if len(ansi.strip_style(divider_char)) != 1:
                 raise TypeError("Divider character must be exactly one character long")
 
             divider_char_width = ansi.style_aware_wcswidth(divider_char)
             if divider_char_width == -1:
-                raise (ValueError("Divider character is an unprintable character"))
+                raise ValueError("Divider character is an unprintable character")
 
         super().__init__(cols, tab_width=tab_width)
         self.divider_char = divider_char
 
     @classmethod
-    def base_width(cls, num_cols: int) -> int:
+    def base_width(cls, num_cols: int, *, column_spacing: int = 2) -> int:
         """
         Utility method to calculate the display width required for a table before data is added to it.
         This is useful when determining how wide to make your columns to have a table be a specific width.
 
         :param num_cols: how many columns the table will have
+        :param column_spacing: how many spaces to place between columns. Defaults to 2.
         :return: base width
+        :raises: ValueError if column_spacing is less than 0
         :raises: ValueError if num_cols is less than 1
         """
         if num_cols < 1:
@@ -571,14 +585,14 @@ class SimpleTable(TableCreator):
         data_str = SPACE
         data_width = ansi.style_aware_wcswidth(data_str) * num_cols
 
-        tbl = cls([Column(data_str)] * num_cols)
+        tbl = cls([Column(data_str)] * num_cols, column_spacing=column_spacing)
         data_row = tbl.generate_data_row([data_str] * num_cols)
 
         return ansi.style_aware_wcswidth(data_row) - data_width
 
     def total_width(self) -> int:
         """Calculate the total display width of this table"""
-        base_width = self.base_width(len(self.cols))
+        base_width = self.base_width(len(self.cols), column_spacing=ansi.style_aware_wcswidth(self.inter_cell))
         data_width = sum(col.width for col in self.cols)
         return base_width + data_width
 
@@ -587,7 +601,7 @@ class SimpleTable(TableCreator):
         header_buf = io.StringIO()
 
         # Create the header labels
-        header = self.generate_row(inter_cell=self.INTER_CELL)
+        header = self.generate_row(inter_cell=self.inter_cell)
         header_buf.write(header)
 
         # Create the divider if necessary
@@ -611,10 +625,9 @@ class SimpleTable(TableCreator):
         :param row_data: data with an entry for each column in the row
         :return: data row string
         """
-        return self.generate_row(row_data=row_data, inter_cell=self.INTER_CELL)
+        return self.generate_row(row_data=row_data, inter_cell=self.inter_cell)
 
-    def generate_table(self, table_data: Sequence[Sequence[Any]], *,
-                       include_header: bool = True, row_spacing: int = 1) -> str:
+    def generate_table(self, table_data: Sequence[Sequence[Any]], *, include_header: bool = True, row_spacing: int = 1) -> str:
         """
         Generate a table from a data set
 
@@ -653,8 +666,8 @@ class BorderedTable(TableCreator):
     Implementation of TableCreator which generates a table with borders around the table and between rows. Borders
     between columns can also be toggled. This class can be used to create the whole table at once or one row at a time.
     """
-    def __init__(self, cols: Sequence[Column], *, tab_width: int = 4,
-                 column_borders: bool = True, padding: int = 1) -> None:
+
+    def __init__(self, cols: Sequence[Column], *, tab_width: int = 4, column_borders: bool = True, padding: int = 1) -> None:
         """
         BorderedTable initializer
 
@@ -704,7 +717,7 @@ class BorderedTable(TableCreator):
         data_width = sum(col.width for col in self.cols)
         return base_width + data_width
 
-    def generate_table_top_border(self):
+    def generate_table_top_border(self) -> str:
         """Generate a border which appears at the top of the header and data section"""
         pre_line = '╔' + self.padding * '═'
 
@@ -715,10 +728,11 @@ class BorderedTable(TableCreator):
 
         post_line = self.padding * '═' + '╗'
 
-        return self.generate_row(row_data=self.empty_data, fill_char='═', pre_line=pre_line,
-                                 inter_cell=inter_cell, post_line=post_line)
+        return self.generate_row(
+            row_data=self.empty_data, fill_char='═', pre_line=pre_line, inter_cell=inter_cell, post_line=post_line
+        )
 
-    def generate_header_bottom_border(self):
+    def generate_header_bottom_border(self) -> str:
         """Generate a border which appears at the bottom of the header"""
         pre_line = '╠' + self.padding * '═'
 
@@ -729,10 +743,11 @@ class BorderedTable(TableCreator):
 
         post_line = self.padding * '═' + '╣'
 
-        return self.generate_row(row_data=self.empty_data, fill_char='═', pre_line=pre_line,
-                                 inter_cell=inter_cell, post_line=post_line)
+        return self.generate_row(
+            row_data=self.empty_data, fill_char='═', pre_line=pre_line, inter_cell=inter_cell, post_line=post_line
+        )
 
-    def generate_row_bottom_border(self):
+    def generate_row_bottom_border(self) -> str:
         """Generate a border which appears at the bottom of rows"""
         pre_line = '╟' + self.padding * '─'
 
@@ -743,10 +758,11 @@ class BorderedTable(TableCreator):
 
         post_line = self.padding * '─' + '╢'
 
-        return self.generate_row(row_data=self.empty_data, fill_char='─', pre_line=pre_line,
-                                 inter_cell=inter_cell, post_line=post_line)
+        return self.generate_row(
+            row_data=self.empty_data, fill_char='─', pre_line=pre_line, inter_cell=inter_cell, post_line=post_line
+        )
 
-    def generate_table_bottom_border(self):
+    def generate_table_bottom_border(self) -> str:
         """Generate a border which appears at the bottom of the table"""
         pre_line = '╚' + self.padding * '═'
 
@@ -757,8 +773,9 @@ class BorderedTable(TableCreator):
 
         post_line = self.padding * '═' + '╝'
 
-        return self.generate_row(row_data=self.empty_data, fill_char='═', pre_line=pre_line,
-                                 inter_cell=inter_cell, post_line=post_line)
+        return self.generate_row(
+            row_data=self.empty_data, fill_char='═', pre_line=pre_line, inter_cell=inter_cell, post_line=post_line
+        )
 
     def generate_header(self) -> str:
         """Generate table header"""
@@ -837,8 +854,17 @@ class AlternatingTable(BorderedTable):
     Implementation of BorderedTable which uses background colors to distinguish between rows instead of row border
     lines. This class can be used to create the whole table at once or one row at a time.
     """
-    def __init__(self, cols: Sequence[Column], *, tab_width: int = 4, column_borders: bool = True, padding: int = 1,
-                 bg_odd: Optional[ansi.bg] = None, bg_even: Optional[ansi.bg] = ansi.bg.bright_black) -> None:
+
+    def __init__(
+        self,
+        cols: Sequence[Column],
+        *,
+        tab_width: int = 4,
+        column_borders: bool = True,
+        padding: int = 1,
+        bg_odd: Optional[ansi.bg] = None,
+        bg_even: Optional[ansi.bg] = ansi.bg.bright_black,
+    ) -> None:
         """
         AlternatingTable initializer
 
@@ -896,8 +922,9 @@ class AlternatingTable(BorderedTable):
         # Apply appropriate background color to data, but don't change the original
         to_display = [self._apply_bg_color(col) for col in row_data]
 
-        row = self.generate_row(row_data=to_display, fill_char=fill_char, pre_line=pre_line,
-                                inter_cell=inter_cell, post_line=post_line)
+        row = self.generate_row(
+            row_data=to_display, fill_char=fill_char, pre_line=pre_line, inter_cell=inter_cell, post_line=post_line
+        )
         self.row_num += 1
         return row
 
